/*
 * (c) Copyright 2002 - 2004, 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "textmenu.h"
#include "textwin.h"
#include "keyinput.h"
#include "elog.h"

#define MAXWIDTH 80

#warning round floating point values (remove rounding from textgui.c)

struct scrollstring {
    int x;
    int id;
    char *str;
};

static int xsize, ysize;
static char *spaces = NULL;
static struct tw_screen *stored_screen = NULL;
static struct scrollstring *scrollstrings = NULL;
static bool_t *yesno_answer = NULL;

static void
state_yesno(int action)
{
    switch (action) {
    case KI_KEY_LEFT:
	*yesno_answer = true;
	tw_restore_screen(stored_screen);
        tw_delete_screen(stored_screen);
        tw_refresh();
	textgui_pop_state();
	break;
    case ACTION_EXIT:
    case KI_KEY_RIGHT:
	*yesno_answer = false;
	tw_restore_screen(stored_screen);
        tw_delete_screen(stored_screen);
        tw_refresh();
	textgui_pop_state();
	break;
    default:
	break;
    }
}

static void
state_message(int action)
{
    elog(LOGDEBUG, "action: %d", action);
    switch (action) {
    case ACTION_EXIT:
    case KI_KEY_LEFT:
	tw_restore_screen(stored_screen);
        tw_delete_screen(stored_screen);
        tw_refresh();
	textgui_pop_state();
	break;
    default:
	break;
    }
}

static void
print_entry(struct textmenu_entry *e,
	    int x,
	    int y,
            bool_t edit_mode)
{
    char str[xsize + 1], c, *p, *p1, *p2, strentry[MAXWIDTH + 1];
    int n, slen, pos, direction;
    double r;
    bool_t b;

    memset(strentry, ' ', MAXWIDTH);
    strentry[MAXWIDTH] = '\0';
    str[xsize] = '\0';
    switch (e->type) {
    default:
    case TEXTMENU_ENTRY_SUBSTATE:
    case TEXTMENU_ENTRY_NONE:
	strncpy(strentry, e->str, MAXWIDTH);
	break;
    case TEXTMENU_ENTRY_ASK:
	if ((p = strchr(e->str, '|')) == NULL) {
	    p = strchr(e->str, '?') + 1;
	}
	c = p[0];
	p[0] = '\0';
	strncpy(strentry, e->str, MAXWIDTH - 1);
	p[0] = c;
	break;
    case TEXTMENU_ENTRY_STRING:
	p = strchr(e->str, '[');
	*p = '\0';
        p1 = strchr(&p[1], ']');
        if (strlen(e->str) + strlen(e->value.string) + strlen(p1) + 1 + x >
            xsize && edit_mode)
        {
            x = 0;
            snprintf(strentry, MAXWIDTH, "%s",  e->value.string);
        } else {
            snprintf(strentry, MAXWIDTH, "%s[%s%s", e->str, e->value.string,
                     p1);
        }
	*p = '[';
	break;
    case TEXTMENU_ENTRY_REAL:
    case TEXTMENU_ENTRY_INTEGER:
	p2 = strchr(e->str, '[');
	p = strchr(p2, ',');
	*p = '\0';
	p1 = strchr(p + 1, ',');
	p1 = strchr(p1 + 1, ',');
	p1 = strchr(p1 + 1, ',');
	p1 = strchr(p1 + 1, ',');
	b = false;
	if (p1 != NULL) {
	    if (e->type == TEXTMENU_ENTRY_REAL) {
                r = atof(p1 + 1);
                b = (fabs(r * 0.999) <= fabs(*e->value.real) &&
                     fabs(r * 1.001) >= fabs(*e->value.real));
	    } else {
		b = *e->value.integer == atof(p1 + 1);
	    }
	}
	if (b) {
	    *p2 = '\0';
	    snprintf(strentry, MAXWIDTH, "%s[%s", e->str,
		     strchr(p1 + 1, ',') + 1);
	    *p2 = '[';
	} else {
	    if (e->type == TEXTMENU_ENTRY_REAL) {
		snprintf(str, xsize, e->str, *e->value.real);
	    } else {
		snprintf(str, xsize, e->str, *e->value.integer);
	    }
	    p1 = strchr(p + 1, ']');
	    snprintf(strentry, MAXWIDTH, "%s%s", str, p1);
	}
	*p = ',';
	break;
    case TEXTMENU_ENTRY_CYCLIC:
	p = strchr(e->str, '[');
	c = p[1];
	p[1] = '\0';
	strncpy(strentry, e->str, MAXWIDTH);
	p[1] = c;	
	p++;
	for (n = 0; n < *e->value.cyclic; n++) {
	    if ((p = strchr(p, '/')) == NULL) {
		break;
	    }
	    p++;
	}
	if (p == NULL || (p1 = strchr(p, ']')) == NULL) {
	    /* cyclic value out of range, reset and redo */
	    *e->value.cyclic = 0;
	    print_entry(e, x, y, false);
	    return;
	}
	p2 = strchr(p, '/');
	if (p2 == NULL || p2 > p1) {
	    slen = strlen(strentry);
	    strncpy(&strentry[slen], p, MAXWIDTH - slen);
	} else {
	    *p2 = '\0';
	    slen = strlen(strentry);
	    strncpy(&strentry[slen], p, MAXWIDTH - slen);
	    slen = strlen(strentry);
	    *p2 = '/';
	    strncpy(&strentry[slen], p1, MAXWIDTH - slen);
	}
	break;
    }
    if ((slen = strlen(strentry)) > xsize - x) {
	pos = tw_get_scrollstring_pos(scrollstrings[y].id, &direction);
	if (scrollstrings[y].id != -1) {
	    if (strcmp(scrollstrings[y].str, strentry) == 0 && pos != -1) {
		return;
	    }
	    free(scrollstrings[y].str);
	    tw_del_scrollstring(scrollstrings[y].id);
	}
	scrollstrings[y].id = tw_add_scrollstring(x, y, strentry, xsize - x,
						  pos == -1 ? 0 : pos,
						  direction, false);
	scrollstrings[y].str = strdup(strentry);
        scrollstrings[y].x = x;
    } else {
	if (scrollstrings[y].id != -1) {
	    tw_del_scrollstring(scrollstrings[y].id);
	    free(scrollstrings[y].str);
	    scrollstrings[y].id = -1;
	}
	tw_putstring(x, y, strentry);
	tw_putstring(x + slen, y, spaces);
    }
}

static void
print_menu(struct textmenu_state *state)
{
    int startpos, selectpos, n, slen, o;
    bool_t new_hdr;

    startpos = (state->pos / ysize) * ysize;
    selectpos = state->pos % ysize;
    o = 0;
    if (state->heading != NULL) {
	if (startpos == 0) {
	    if (selectpos == ysize - 1) {
		startpos = ysize - 1;
		selectpos = 0;
	    } else {
		slen = strlen(state->heading);
                new_hdr = false;
                if (scrollstrings[0].id != -1 &&
                    scrollstrings[0].x == 0 &&
                    strcmp(state->heading, scrollstrings[0].str) == 0)
                {
                    if (tw_get_scrollstring_pos(scrollstrings[0].id, &n)
                        == -1)
                    {
                        free(scrollstrings[0].str);
                        scrollstrings[0].id = -1;
                        new_hdr = true;
                    }
                } else {
                    new_hdr = true;
                }
                if (new_hdr) {
                    if (scrollstrings[0].id != -1) {
                        tw_del_scrollstring(scrollstrings[0].id);
                        free(scrollstrings[0].str);
                        scrollstrings[0].id = -1;
                    }
                    if (slen <= xsize) {
                        tw_putstring(0, 0, state->heading);
                        tw_putstring(slen, 0, spaces);
                    } else {
                        scrollstrings[0].id =
                            tw_add_scrollstring(0, 0, state->heading, xsize,
                                                0, 1, false);
                        scrollstrings[0].str = strdup(state->heading);
                        scrollstrings[0].x = 0;
                    }
                }
		o = 1;
	    }
	} else {
	    startpos -= 1;
	    selectpos += 1;
	    if (selectpos == ysize) {
		selectpos = 0;
		startpos += ysize;
	    }
	}
    }

    /* place the arrows */
    for (n = 0; n < ysize - o; n++) {
	tw_putchar(1, n + o, n == selectpos ? TWC_RIGHT : ' ');
    }
    tw_putchar(0, o, startpos != 0 ? TWC_UP : ' ');
    tw_putchar(0, ysize - 1, state->n_entries - startpos + o > ysize ?
	       TWC_DOWN : ' ');

    for (n = 0; n < ysize - o; n++) {
	if (startpos + n < state->n_entries) {
	    print_entry(&state->entries[startpos + n], 2, n + o, false);
	} else {
            if (scrollstrings[n + o].id != -1) {
                tw_del_scrollstring(scrollstrings[n + o].id);
                free(scrollstrings[n + o].str);
                scrollstrings[n + o].id = -1;
            }
	    tw_putstring(2, n + o, spaces);
	}
    }
    tw_refresh();
}

struct textmenu_state *
textmenu_state_new(const char heading[],
		   void *entries[])
{
    struct textmenu_state *state;
    char *s, *p;
    int n = 0;

    xsize = tw_xsize();
    ysize = tw_ysize();
    if (spaces == NULL) {
	spaces = malloc(xsize + 1);
	for (n = 0; n < xsize; n++) {
	    spaces[n] = ' ';
	}
	spaces[n] = '\0';
	n = 0;
    }
    if (scrollstrings == NULL) {
	scrollstrings = malloc(ysize * sizeof(struct scrollstring));
	memset(scrollstrings, 0xFF, ysize * sizeof(struct scrollstring));
    }
    
    state = malloc(sizeof(struct textmenu_state));
    memset(state, 0, sizeof(struct textmenu_state));
    if (heading != NULL) {
	state->heading = strdup(heading);
    }
    while (entries[n] != NULL) {
	state->entries = realloc(state->entries, (state->n_entries + 1) *
				 sizeof(struct textmenu_entry));
	memset(&state->entries[state->n_entries], 0,
	       sizeof(struct textmenu_entry));
	state->entries[state->n_entries].str =
	    strdup((const char *)entries[n]);
	s = state->entries[state->n_entries].str;
	if (strstr(s, "...") != NULL) {
	    state->entries[state->n_entries].type =
		TEXTMENU_ENTRY_SUBSTATE;
	    state->entries[state->n_entries].value.substate =
		(void (*)(int))entries[++n];
	} else if ((p = strchr(s, '[')) != NULL &&
		   (p = strchr(p, '/')) != NULL &&
		   strchr(p, ']') != NULL)
	{
	    state->entries[state->n_entries].type =
		TEXTMENU_ENTRY_CYCLIC;
	    state->entries[state->n_entries].value.cyclic =
		(int *)entries[++n];
	} else if ((p = strchr(s, '[')) != NULL &&
		   (p = strchr(p, '%')) != NULL &&
		   strchr(p, ']') != NULL)
	{
	    while (true) {
		p++;                
		if (*p == 'f') {
		    state->entries[state->n_entries].type = TEXTMENU_ENTRY_REAL;
		    state->entries[state->n_entries].value.real =
			(double *)entries[++n];
		    break;
		} else if (*p == 'd') {
		    state->entries[state->n_entries].type =
			TEXTMENU_ENTRY_INTEGER;
		    state->entries[state->n_entries].value.integer =
			(int *)entries[++n];
		    break;
		} else if (*p == 's') {
		    state->entries[state->n_entries].type =
			TEXTMENU_ENTRY_STRING;
		    state->entries[state->n_entries].value.string =
			(char *)entries[++n];
		    break;
		}
	    }
	} else if (strchr(s, '?') != NULL || strchr(s, '|') != NULL) {
	    state->entries[state->n_entries].type =
		TEXTMENU_ENTRY_ASK;
	} else {
	    state->entries[state->n_entries].type =
		TEXTMENU_ENTRY_NONE;
	}
	state->n_entries++;
	n++;
    }
    return state;
}

void
textmenu_state_delete(struct textmenu_state *state)
{
    int n;
    
    if (state == NULL) {
	return;
    }
    free(state->heading);
    for (n = 0; n < state->n_entries; n++) {
	free(state->entries[n].str);
    }
    free(state);
}

void
textmenu_action(struct textmenu_state *state,
		int *action)
{
    static int substate = TEXTMENU_ENTRY_NONE;
    static double origval, multiplier, multiplier_max, multiplier_min;
    static double value_min, value_max;
    static int update_line, estrpos, maxstrlen;
    static char *origstring = NULL;
    static const char *estr = "AC abcdefghijklmnopqrstuvwxyz-_. 0123456789 CA";
    char *p, str[xsize + 100], str2[xsize + 100];
    int n, line, slen;
    double r, newr;

    if (state == NULL) {
        return;
    }
    if (state->pos >= state->n_entries) {
        state->pos = 0;
    }
    
    if (*action == ACTION_EXIT) {
	return;
    }
    
    switch (substate) {
    case TEXTMENU_ENTRY_ASK:
	switch (*action) {
	case KI_KEY_LEFT:
	    substate = TEXTMENU_ENTRY_NONE;
	    *action = state->pos;
	    tw_clear();
	    print_menu(state);
	    break;
	case KI_KEY_RIGHT:
	    substate = TEXTMENU_ENTRY_NONE;
	    *action = ACTION_IGNORE;
	    tw_clear();
	    print_menu(state);
	    break;
	default:
	    *action = ACTION_IGNORE;
	    break;
	}
	return;
    case TEXTMENU_ENTRY_STRING:
        switch (*action) {
        case KI_KEY_LEFT:
            slen = strlen(state->entries[state->pos].value.string);
            if (slen > 0) {
                state->entries[state->pos].value.string[slen-1] = '\0';
            }
            break;
        case KI_KEY_RIGHT:
            switch (estr[estrpos]) {
            case 'A':
                substate = TEXTMENU_ENTRY_NONE;
                if (strcmp(origstring,
                           state->entries[state->pos].value.string) != 0)
                {
                    *action = state->pos;
                } else {
                    *action = ACTION_IGNORE;
                }
                tw_clear();
                print_menu(state);
                return;
            case 'C':
                substate = TEXTMENU_ENTRY_NONE;
                strcpy(state->entries[state->pos].value.string, origstring);
                *action = ACTION_IGNORE;
                tw_clear();
                print_menu(state);
                return;
            default:
                slen = strlen(state->entries[state->pos].value.string);
                if (slen < maxstrlen) {
                    state->entries[state->pos].value.string[slen] =
                        estr[estrpos];
                    state->entries[state->pos].value.string[slen+1] = '\0';
                }
                break;
            }
            break;
        case KI_KEY_DOWN:
            if (estrpos < strlen(estr) - 1) {
                estrpos++;
            }
            break;
        case KI_KEY_UP:
            if (estrpos > 0) {
                estrpos--;
            }
            break;
        case ACTION_REFRESH:
            break;
        default:
            *action = ACTION_IGNORE;
            return;
        }
        tw_clear();
        print_entry(&state->entries[state->pos], 2, update_line, true);
        line = (update_line == ysize - 1) ? ysize - 2 : update_line + 1;
        slen = strlen(estr);
        for (n = -2; n <= 2; n++) {
            if (estrpos + n < 0 || estrpos + n >= slen) {
                str[n+2] = ' ';                
            } else if (n == 0) {
                str[n+2] = toupper(estr[estrpos+n]);
            } else {
                str[n+2] = estr[estrpos+n];
            }
        }
        str[5] = '\0';
        switch (estr[estrpos]) {
        case 'A':
            strcpy(str2, "accept");
            break;
        case 'C':
            strcpy(str2, "cancel");
            break;
        default:
            if (strlen(state->entries[state->pos].value.string) == maxstrlen) {
                strcpy(str2, "MAXLEN");
            } else {
                strcpy(str2, "choose");
            }
            break;
        }
        tw_printf(0, line, "%c%s%c %c%s %cdel",
                  TWC_UP, str, TWC_DOWN, TWC_RIGHT, str2, TWC_LEFT);
        tw_refresh();
        *action = ACTION_IGNORE;
        return;
    case TEXTMENU_ENTRY_REAL:
    case TEXTMENU_ENTRY_INTEGER:
	if (substate == TEXTMENU_ENTRY_REAL) {
	    r = *state->entries[state->pos].value.real;
	} else {
	    r = (double)*state->entries[state->pos].value.integer;
	}
	switch (*action) {
	case KI_KEY_LEFT:
	    substate = TEXTMENU_ENTRY_NONE;
	    if (r != origval) {
		*action = state->pos;
	    } else {
		*action = ACTION_IGNORE;
	    }
	    tw_clear();
	    print_menu(state);
	    return;
	case KI_KEY_RIGHT:
	    multiplier *= 10.0;
	    if (multiplier > multiplier_max) {
		multiplier = multiplier_min;
	    }
	    if (multiplier < 0.011 && multiplier > 0.009) {
		multiplier = 0.01;
		strcpy(str, "0.01");
	    } else if (multiplier < 0.11 && multiplier > 0.09) {
		multiplier = 0.1;
		strcpy(str, "0.1 ");
	    } else if (multiplier < 1.01 && multiplier > 0.99) {
		multiplier = 1.0;
		if (substate == TEXTMENU_ENTRY_REAL) {
		    strcpy(str, "1.0 ");
		} else {
		    strcpy(str, "1   ");
		}
	    } else if (multiplier < 10.1 && multiplier > 9.9) {
                multiplier = 10.0;
		strcpy(str, "10  ");
	    } else if (multiplier < 100.1 && multiplier > 99.9) {
                multiplier = 100.0;
		strcpy(str, "100 ");
	    } else if (multiplier < 1000.1 && multiplier > 999.9) {
                multiplier = 1000.0;
		strcpy(str, "1000");
	    } else {
		elog(LOGERROR, "unexpected multiplier value %f", multiplier); 
		strcpy(str, "ERR");
	    }
	    line = (update_line == ysize - 1) ? ysize - 2 : update_line + 1;
	    tw_printf(0, line, "  %cok %c*%s %c+ %c-", TWC_LEFT, TWC_RIGHT, str,
		      TWC_UP, TWC_DOWN);
	    tw_refresh();
	    *action = ACTION_IGNORE;
	    break;
	case KI_KEY_UP:
	    if (r <= value_max - multiplier) {
		if (substate == TEXTMENU_ENTRY_REAL) {
		    *state->entries[state->pos].value.real += multiplier;
		    newr = *state->entries[state->pos].value.real;
		} else {
		    *state->entries[state->pos].value.integer +=
			(int)multiplier;
		    newr = *state->entries[state->pos].value.integer;
		}
	    } else {
		if (substate == TEXTMENU_ENTRY_REAL) {
		    newr = *state->entries[state->pos].value.real = value_max;
		} else {
		    *state->entries[state->pos].value.integer = (int)value_max;
		    newr = value_max;
		}
	    }
	    print_entry(&state->entries[state->pos], 2, update_line, true);
	    tw_refresh();
	    if (newr != r) {
		*action = state->pos;
	    } else {
		*action = ACTION_IGNORE;
            }
	    break;
	case KI_KEY_DOWN:
	    if (r >= value_min + multiplier) {
		if (substate == TEXTMENU_ENTRY_REAL) {
		    *state->entries[state->pos].value.real -= multiplier;
		    newr = *state->entries[state->pos].value.real;
		} else {
		    *state->entries[state->pos].value.integer -=
			(int)multiplier;
		    newr = *state->entries[state->pos].value.integer;
		}
	    } else {
		if (substate == TEXTMENU_ENTRY_REAL) {
		    newr = *state->entries[state->pos].value.real = value_min;
		} else {
		    *state->entries[state->pos].value.integer = (int)value_min;
		    newr = value_min;
		}
	    }
	    print_entry(&state->entries[state->pos], 2, update_line, true);
	    tw_refresh();
	    if (newr != r) {
		*action = state->pos;
	    } else {
		*action = ACTION_IGNORE;
            }
	    break;
        case ACTION_REFRESH:
	    print_entry(&state->entries[state->pos], 2, update_line, true);
	    tw_refresh();
            break;
	default:
	    *action = ACTION_IGNORE;
	    break;
	}
	return;
    case TEXTMENU_ENTRY_NONE:
	break;
    }

    switch (*action) {
    case KI_KEY_UP:
	*action = ACTION_IGNORE;
	if (state->pos == 0) {
	    return;
	}
	state->pos--;
	break;
    case KI_KEY_DOWN:
	*action = ACTION_IGNORE;
	if (state->pos == state->n_entries - 1) {
	    return;
	}
	state->pos++;
	break;
    case KI_KEY_LEFT:
	*action = ACTION_EXIT;
	break;
    case KI_KEY_RIGHT:
	switch (state->entries[state->pos].type) {
	default:
	case TEXTMENU_ENTRY_NONE:
	    *action = state->pos;
	    return;
	case TEXTMENU_ENTRY_SUBSTATE:
	    *action = ACTION_IGNORE;
	    textgui_push_state(state->entries[state->pos].value.substate);
	    return;
	case TEXTMENU_ENTRY_ASK:
	    if ((p = strchr(state->entries[state->pos].str, '|')) == NULL) {
		p = strchr(state->entries[state->pos].str, '?');
	    }
	    p = &p[1];
	    if (strcmp(p, "ask") == 0) {
                substate = TEXTMENU_ENTRY_ASK;
		*action = ACTION_IGNORE;
		line = 0;
		for (n = 0; n < ysize; n++) {
		    if (tw_getchar(1, n) != TWC_RIGHT) {
                        tw_del_scrollstring_line(n);
			tw_putstring(0, n, spaces);
		    } else {
			line = n;
		    }
		}
		tw_putchar(1, line, ' ');
		line = (line == ysize - 1) ? ysize - 2 : line + 1;
		tw_printf(0, line, "  %cyes   %cno", TWC_LEFT, TWC_RIGHT);
		tw_refresh();
		return;
	    } else if (*p != '\0') {
		substate = TEXTMENU_ENTRY_ASK;
		*action = ACTION_IGNORE;
		tw_clear();
		tw_putstring(0, 0, p);
		tw_printf(0, 1, "  %cyes   %cno", TWC_LEFT, TWC_RIGHT);
		tw_refresh();
		return;
	    } else {
		*action = state->pos;
	    }
	    break;
        case TEXTMENU_ENTRY_STRING:
            substate = TEXTMENU_ENTRY_STRING;
	    for (n = 0; n < ysize; n++) {
		if (tw_getchar(1, n) == TWC_RIGHT) {
		    update_line = n;
		}
	    }
            free(origstring);
            origstring = strdup(state->entries[state->pos].value.string);
            estrpos = 3;
	    p = strchr(state->entries[state->pos].str, '[');
	    p = strchr(p, ',') + 1;
            maxstrlen = atoi(p);
	    *action = ACTION_REFRESH;
	    textmenu_action(state, action);
            *action = ACTION_IGNORE;
            return;
	case TEXTMENU_ENTRY_REAL:
	case TEXTMENU_ENTRY_INTEGER:
	    substate = state->entries[state->pos].type;
	    for (n = 0; n < ysize; n++) {
		if (tw_getchar(1, n) == TWC_RIGHT) {
		    update_line = n;
		}
	    }
            tw_clear();
	    print_entry(&state->entries[state->pos], 2, update_line, true);
            tw_refresh();
	    if (*action == TEXTMENU_ENTRY_REAL) {
		origval = *state->entries[state->pos].value.real;
	    } else {
		origval = (double)*state->entries[state->pos].value.integer;
	    }
	    p = strchr(state->entries[state->pos].str, '[');
	    p = strchr(p, ',') + 1;
	    multiplier_min = atof(p);
	    p = strchr(p, ',') + 1;
	    multiplier_max = atof(p);
	    p = strchr(p, ',') + 1;
	    value_min = atof(p);
	    p = strchr(p, ',') + 1;
	    value_max = atof(p);
	    multiplier = multiplier_max;
	    *action = KI_KEY_RIGHT;
	    textmenu_action(state, action);
	    *action = ACTION_IGNORE;
	    return;
	case TEXTMENU_ENTRY_CYCLIC:
	    *action = state->pos;
	    *state->entries[state->pos].value.cyclic += 1;
	    break;
	}
	break;
    case ACTION_CHILD_POP:
    case ACTION_INIT:
	tw_clear();	
	break;
    default:
	return;
    case ACTION_REFRESH:
        break;
    }
    print_menu(state);
}

void
textmenu_showpage(int *action,
		  int *pos,
		  const char *lines[],
		  int n_lines)
{
    int n, slen;
    
    switch (*action) {
    case ACTION_IGNORE:
	return;
    case ACTION_INIT:
	*pos = 0;
	break;
    case KI_KEY_UP:
	if (*pos >= ysize) {
	    *pos -= ysize;
	}
	*action = ACTION_IGNORE;
	break;
    case KI_KEY_DOWN:	
	if (*pos + ysize < n_lines) {
	    *pos += ysize;
	}
	*action = ACTION_IGNORE;
	break;
    case KI_KEY_LEFT:
	*action = ACTION_EXIT;
	return;
    }
    tw_clear();
    for (n = *pos; n < *pos + ysize && n < n_lines; n++) {
	if ((slen = strlen(lines[n])) > xsize - 1) {
	    tw_add_scrollstring(1, n - (*pos), lines[n], xsize - 1, 0, 1,
				false);
	} else {
	    tw_putstring(1, n - (*pos), lines[n]);
	}
    }
    if (*pos != 0) {
	tw_putchar(0, 0, TWC_UP);
    }
    if (n != n_lines) {
	tw_putchar(0, ysize - 1, TWC_DOWN);
    }
    tw_refresh();
}

void
textmenu_yesno(bool_t *answer,
               const char line1[],
               const char line2[])
{
    stored_screen = tw_store_screen();
    tw_clear();
    tw_putstring(0, 0, line1);
    tw_putstring(0, 1, line2);
    tw_printf(xsize - 8, ysize - 1, "%cyes %cno", TWC_LEFT, TWC_RIGHT);
    tw_refresh();
    yesno_answer = answer;
    textgui_push_state(state_yesno);
}

void
textmenu_message(const char line1[],
		 const char line2[])
{    
    stored_screen = tw_store_screen();
    tw_clear();
    tw_putstring(0, 0, line1);
    tw_putstring(0, 1, line2);
    tw_putstring(xsize - 2, ysize - 1, "ok");
    tw_putchar(xsize - 3, ysize - 1, TWC_LEFT);
    tw_refresh();
    textgui_push_state(state_message);
    elog(LOGDEBUG, "line1: \"%s\"", line1);
    elog(LOGDEBUG, "line2: \"%s\"", line2);
}
