/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _TEXTWIN_H_
#define _TEXTWIN_H_

#include "defs.h"

#define TWC_LEFT  1
#define TWC_RIGHT 2
#define TWC_UP    3
#define TWC_DOWN  4

struct textwin_output {
    /* textwin_output modules don't need to check input parameters, any writing
       outside windows etc will be taken care of before passing to the output */
    
    /* required functions */
    void (*properties)(int xsize_range[2],
		       int ysize_range[2],
		       char direction_chars[4]);
    void (*putchar)(int x,
		    int y,
		    char c);
    /* optional functions */
    bool_t (*init)(int xsize,
		   int ysize);
    void (*refresh)(void);
    void (*clear)(void);
    void (*putstring)(int x,
		       int y,
		       const char s[]);    
    void (*brightness)(int value);
};

bool_t
tw_init(struct textwin_output outputs[],
	int n_outputs,
	int xsize,
	int ysize,
	int (*register_timer)(void *param,
			      void (*callback)(void *param),
			      struct timeval *timeval),
	void (*unregister_timer)(int id));

int
tw_xsize(void);

int
tw_ysize(void);

void
tw_clear(void);

char
tw_getchar(int x,
	   int y);

void
tw_putchar(int x,
	   int y,
	   char c);

void
tw_putstring(int x,
	     int y,
	     const char s[]);

void
tw_refresh(void);

void
tw_printf(int x,
	  int y,
	  const char format[],
	  ...);

int
tw_add_scrollstring(int x,
		    int y,
		    const char s[],
		    int len,
		    int startpos,
		    int direction,
		    bool_t no_reverse);

void
tw_del_scrollstring(int id);

void
tw_del_scrollstring_line(int y);

int
tw_get_scrollstring_pos(int id,
			int *direction);

void
tw_print_scrollstrings(bool_t do_print);

struct tw_screen;

struct tw_screen *
tw_store_screen(void);

void
tw_restore_screen(const struct tw_screen *screen);

void
tw_delete_screen(struct tw_screen *screen);

void
tw_brightness(int value);

#endif
